/*
 * File: tnef.js
 *   extract files from microsoft TNEF format
 *
 * Copyright (C) 2007 Aron Rubin <arubin@atl.lmco.com>
 *
 *
 * About:
 *       scans tnef file and extracts all attachments
 *       attachments are written to their original file-names if possible
 */

// ---------- Tnef Types ----------

const szTRIPLES = 0x0000; //   triples
const szSTRING  = 0x0001; //   string
const szTEXT    = 0x0002; //   text
const szDATE    = 0x0003; //   date
const szSHORT   = 0x0004; //   short
const szLONG    = 0x0005; //   long
const szBYTE    = 0x0006; //   byte
const szWORD    = 0x0007; //   word
const szDWORD   = 0x0008; //   dword
const szMAX     = 0x0009; //   max


// ---------- Tnef Names -----------

// names of all attributes found in TNEF file
const attOWNER                        = 0x0000; //  Owner
const attSENTFOR                      = 0x0001; //  Sent For
const attDELEGATE                     = 0x0002; //  Delegate
const attDATESTART                    = 0x0006; //  Date Start
const attDATEEND                      = 0x0007; //  Date End
const attAIDOWNER                     = 0x0008; //  Owner Appointment ID
const attREQUESTRES                   = 0x0009; //  Response Requested.
const attFROM                         = 0x8000; //  From
const attSUBJECT                      = 0x8004; //  Subject
const attDATESENT                     = 0x8005; //  Date Sent
const attDATERECD                     = 0x8006; //  Date Received
const attMESSAGESTATUS                = 0x8007; //  Message Status
const attMESSAGECLASS                 = 0x8008; //  Message Class
const attMESSAGEID                    = 0x8009; //  Message ID
const attPARENTID                     = 0x800a; //  Parent ID
const attCONVERSATIONID               = 0x800b; //  Conversation ID
const attBODY                         = 0x800c; //  Body
const attPRIORITY                     = 0x800d; //  Priority
const attATTACHDATA                   = 0x800f; //  Attachment Data
const attATTACHTITLE                  = 0x8010; //  Attachment File Name
const attATTACHMETAFILE               = 0x8011; //  Attachment Meta File
const attATTACHCREATEDATE             = 0x8012; //  Attachment Creation Date
const attATTACHMODIFYDATE             = 0x8013; //  Attachment Modification Date
const attDATEMODIFY                   = 0x8020; //  Date Modified
const attATTACHTRANSPORTFILENAME      = 0x9001; //  Attachment Transport Filename
const attATTACHRENDDATA               = 0x9002; //  Attachment Rendering Data
const attMAPIPROPS                    = 0x9003; //  MAPI Properties
const attRECIPTABLE                   = 0x9004; //  Recipients
const attATTACHMENT                   = 0x9005; //  Attachment
const attTNEFVERSION                  = 0x9006; //  TNEF Version
const attOEMCODEPAGE                  = 0x9007; //  OEM Codepage
const attORIGNINALMESSAGECLASS        = 0x9008; //  Original Message Class


// ---------- MAPI Types -----------

// all types found in MAPI TNEF attachment
const szMAPI_UNSPECIFIED    = 0x0000; // MAPI Unspecified
const szMAPI_NULL           = 0x0001; // MAPI null property
const szMAPI_SHORT          = 0x0002; // MAPI short (signed 16 bits)
const szMAPI_INT            = 0x0003; // MAPI integer (signed 32 bits)
const szMAPI_FLOAT          = 0x0004; // MAPI float (4 bytes)
const szMAPI_DOUBLE         = 0x0005; // MAPI double
const szMAPI_CURRENCY       = 0x0006; // MAPI currency (64 bits)
const szMAPI_APPTIME        = 0x0007; // MAPI application time
const szMAPI_ERROR          = 0x000a; // MAPI error (32 bits)
const szMAPI_BOOLEAN        = 0x000b; // MAPI boolean (16 bits)
const szMAPI_OBJECT         = 0x000d; // MAPI embedded object
const szMAPI_INT8BYTE       = 0x0014; // MAPI 8 byte signed int
const szMAPI_STRING         = 0x001e; // MAPI string
const szMAPI_UNICODE_STRING = 0x001f; // MAPI unicode-string (null terminated)
const szMAPI_SYSTIME        = 0x0040; // MAPI time (64 bits)
const szMAPI_CLSID          = 0x0048; // MAPI OLE GUID
const szMAPI_BINARY         = 0x0102; // MAPI binary


// ---------- MAPI Names -----------

// names of some attributes found in TNEF file
const MAPI_PRIORITY = 0x0026;
const MAPI_READ_RECEIPT_REQUESTED = 0x0029;
const MAPI_SUBJECT = 0x0037;
const MAPI_RECEIVED_BY_NAME = 0x0040;
const MAPI_RECEIVED_BY_ADDRTYPE = 0x0075;
const MAPI_RECEIVED_BY_EMAIL_ADDRESS = 0x0076;
const MAPI_TNEF_CORRELATION_KEY = 0x007F;
const MAPI_BODY = 0x1000;
const MAPI_REPORT_TEXT = 0x1001;
const MAPI_RTF_SYNC_BODY_CRC = 0x1006;
const MAPI_RTF_SYNC_BODY_COUNT = 0x1007;
const MAPI_RTF_SYNC_BODY_TAG = 0x1008;
const MAPI_RTF_COMPRESSED = 0x1009;
const MAPI_RTF_SYNC_PREFIX_COUNT = 0x1010;
const MAPI_RTF_SYNC_TRAILING_COUNT = 0x1011;
const MAPI_BODY_HTML = 0x1013;
const MAPI_SMTP_MESSAGE_ID = 0x1035;
const MAPI_EXPLICIT_CONVERSION = 0x0C01;
const MAPI_MESSAGE_TOKEN = 0x0C03;
const MAPI_REPLY_REQUESTED = 0x0C17;
const MAPI_SENDER_NAME = 0x0C1A;
const MAPI_SENDER_EMAIL_ADDRESS = 0x0C1F;
const MAPI_DISPLAY_BCC = 0x0E02;
const MAPI_DISPLAY_CC = 0x0E03;
const MAPI_DISPLAY_TO = 0x0E04;
const MAPI_PARENT_DISPLAY = 0x0E05;
const MAPI_MESSAGE_DELIVERY_TIME = 0x0E06;
const MAPI_MESSAGE_FLAGS = 0x0E07;
const MAPI_MESSAGE_SIZE = 0x0E08;
const MAPI_PARENT_ENTRYID = 0x0E09;
const MAPI_MESSAGE_RECIPIENTS = 0x0E12;
const MAPI_MESSAGE_ATTACHMENTS = 0x0E13;
const MAPI_HASATTACH = 0x0E1B;
const MAPI_BODY_CRC = 0x0E1C;
const MAPI_NORMALIZED_SUBJECT = 0x0E1D;
const MAPI_RTF_IN_SYNC = 0x0E1F;
const MAPI_ATTACH_SIZE = 0x0E20;
const MAPI_ATTACH_NUM = 0x0E21;
const MAPI_PREPROCESS = 0x0E22;
const MAPI_ENTRYID = 0x0FFF;
const MAPI_OBJECT_TYPE = 0x0FFE;
const MAPI_EMAIL_ADDRESS = 0x3003;
const MAPI_CREATION_TIME = 0x3007;
const MAPI_CONTAINER_FLAGS = 0x3600;
const MAPI_FOLDER_TYPE = 0x3601;
const MAPI_CONTENT_COUNT = 0x3602;
const MAPI_CONTENT_UNREAD = 0x3603;
const MAPI_CONTAINER_HIERARCHY = 0x360E;
const MAPI_CONTAINER_CONTENTS = 0x360F;
const MAPI_FOLDER_ASSOCIATED_CONTENTS = 0x3610;
const MAPI_CONTAINER_CLASS = 0x3613;
const MAPI_CONTAINER_MODIFY_VERSION = 0x3614;
const MAPI_ATTACHMENT_X400_PARAMETERS = 0x3700;
const MAPI_ATTACH_DATA_OBJ = 0x3701;
const MAPI_ATTACH_ENCODING = 0x3702;
const MAPI_ATTACH_EXTENSION = 0x3703;
const MAPI_ATTACH_FILENAME = 0x3704;
const MAPI_ATTACH_METHOD = 0x3705;
const MAPI_ATTACH_LONG_FILENAME = 0x3707;
const MAPI_ATTACH_PATHNAME = 0x3708;
const MAPI_ATTACH_RENDERING = 0x3709;
const MAPI_ATTACH_TAG = 0x370A;
const MAPI_ATTACH_TRANSPORT_NAME = 0x370C;
const MAPI_ATTACH_LONG_PATHNAME = 0x370D;
const MAPI_ATTACH_MIME_TAG = 0x370E;
const MAPI_ATTACH_ADDITIONAL_INFO = 0x370F;
const MAPI_ATTACH_MIME_SEQUENCE = 0x3710;
const MAPI_ATTACH_CONTENT_ID = 0x3712;
const MAPI_ATTACH_CONTENT_LOCATION = 0x3713;
const MAPI_ATTACH_FLAGS = 0x3714;
const MAPI_DISPLAY_TYPE = 0x3900;
const MAPI_TEMPLATEID = 0x3902;


// ---------- Util ----------

function assert( condition, caller_arguments ) {
  var caller_arguments = new Array();
  // Function.arguments is deprecated, but convenient if it works
  if( arguments.length < 2 && assert.caller.arguments )
    caller_arguments = assert.caller.arguments;

  if( !condition )
    alert( "Assertion failed in " + assert.caller.name + "( )" );
}

function dump( msg ) {
  var cs = Components.classes["@mozilla.org/consoleservice;1"]
            .getService( Components.interfaces.nsIConsoleService );
  cs.logStringMessage( msg );
}

function GETINT64( p, i ) {
  if( arguments.length < 2 )
    i = 0;

  var val = p.charCodeAt( i + 7 );
  for( var j = 6; j >= 0; j-- )
    val = val*256 + p.charCodeAt( i + j );
}

function GETINT32( p, i ) {
  if( arguments.length < 2 )
    i = 0;
  if( typeof(p) == "string" )
    return( ((p.charCodeAt(i + 3)*256 + p.charCodeAt(i + 2))*256 + p.charCodeAt(i + 1))*256 + p.charCodeAt(i) );
  else
    return( ((p[i + 3]*256 + p[i + 2])*256 + p[i + 1])*256 + p[i] );
}

function GETINT16( p, i ) {
  if( arguments.length < 2 )
    i = 0;
  if( typeof(p) == "string" )
    return( p.charCodeAt(i) + p.charCodeAt(i + 1)*256 );
  else
    return( p[i] + p[i + 1]*256 );
}

function to_hex( x, len ) {
  var range = "0123456789abcdef";
  var tmp = Math.floor( x );
  var hex = "";

  len = len*2;

  while( (!len && tmp) || len ) {
    hex = range[tmp&0x0f] + hex;
    tmp = Math.floor( tmp/16 );
    if( len )
      len--;
  }

  return( hex );
}

function unicode_to_utf8( buf ) {
  var ipos = 0;
  var opos = 0;
  var utf8 = new String( "" ); // won't get any longer than 3*buf.length/2 + 1

  for( ipos = 0; ipos < buf.length - 1; ipos += 2 ) {
    var c = GETINT16( buf, ipos );
    if( c <= 0x007f )	{
      utf8 += String.fromCharCode( 0x00 | ((c & 0x007f) >> 0) );
    } else if( c < 0x07ff ) {
      utf8 += String.fromCharCode( 0xc0 | ((c & 0x07c0) >> 6) );
      utf8 += String.fromCharCode( 0x80 | ((c & 0x003f) >> 0) );
    } else {
      utf8 += String.fromCharCode( 0xe0 | ((c & 0xf000) >> 12) );
      utf8 += String.fromCharCode( 0x80 | ((c & 0x0fc0) >> 6) );
      utf8 += String.fromCharCode( 0x80 | ((c & 0x003f) >> 0) );
    }
  }

  return( utf8 );
}

// ---------- Path ----------

function concat_fname( fname1, fname2 ) {
  var filename = null;

  assert( fname1 || fname2 );

  if( !(fname1 || fname2) )
    return( null );

  if( !fname1 ) {
    filename = fname2;
  } else {
    filename = fname1;

    if( fname2 ) {
      if( filename[filename.length - 1] != '/' && fname2[0] != '/' )
	filename += "/";
    }
    filename += fname2;
  }

  return( filename );
}

function fs_file_exists( fname ) {
  var lf = Components.classes["@mozilla.org/file/local;1"].createInstance();
  lf.initWithPath( fname );
  return( lf.exists() );
}

// finds a filename fname.N where N >= 1 and is not the name of an existing
//   filename.  Assumes that fname does not already have such an extension
function fs_find_free_number( fname ) {
  var newname = "";
  var counter = 1;

  do {
    newname = fname + counter;
    counter++;
  } while( file_exists( newname ) );

  return( newname );
}

function fs_munge_fname( directory, fname ) {
  var file = null;

  // If we were not given a filename make one up
  if( !fname || fname.length == 0 ) {
    var tmp = concat_fname( directory, "tnef-part" );
    //debug_print( "No file name specified, using default.\n" );
    file = find_free_number( tmp );
  } else {
    file = concat_fname( directory, fname );
  }

  return( file );
}


// ---------- File ----------
function TnefFile( pkg ) {
  this.pkg = pkg;
}
TnefFile.prototype = {
  pkg: null,
  name: null,
  len: 0,
  data: null,
  date: Date(),
  mime_type: null,
  position: 0
}

function tnef_file_name_used( fname, files ) {
  var i = 0;

  if( !files )
    return( 0 );

  for( i = 0; i < files.length; i++ )
    if( files[i].name == fname )
      return( 1 );
  return( 0 );
}


function tnef_file_munge_fname( fname, files ) {
  var file = null;
  var count = 0;

  // If we were not given a filename make one up
  if( !fname || fname.length == 0 ) {
    //debug_print( "No file name specified, using default.\n" );
    fname = "tnef-part-";
    do {
      file = fname + count;
    } while( tnef_file_name_used( file, files ) );
  } else {
    file = fname;
    while( tnef_file_name_used( file, files ) )
      file = fname + count;
  }

  return( file );
}


function tnef_file_write_stream( file, outstrm ) {
  outstrm.write( file.data, file.len );
}


function tnef_file_notify( file, listener, final ) {
  dump( "notifying listener" );

  if( !listener )
    return;
  if( file.data || final ) {
    if( file.position == 0 && listener.onTnefStart )
      listener.onTnefStart( file.name, file.mime_type, file.len, file.date );
    if( file.data ) {
      if( listener.onTnefData )
	listener.onTnefData( file.position, file.data );
      file.position += file.data.length;
    }
    if( final && listener.onTnefEnd )
      listener.onTnefEnd( );
  }
}


function tnef_file_add_mapi_attrs( file, files, attrs ) {
  var i = 0;

  for( i = 0; attrs[i]; i++ ) {
    if( attrs[i].num_values ) {
      switch( attrs[i].name ) {
      case MAPI_ATTACH_LONG_FILENAME:
	file.name = tnef_file_munge_fname( attrs[i].values[0], files );
	break;

      case MAPI_ATTACH_DATA_OBJ:
	file.len = attrs[i].values[0].length;
	file.data = attrs[i].values[0];
	break;

      case MAPI_ATTACH_MIME_TAG:
	file.mime_type = attrs[i].values[0];
	break;

      default:
	break;
      }
    }
  }
}

function tnef_file_add_attr( file, files, attr ) {
  assert( file && attr );
  if( !(file && attr) )
    return;

  // we only care about some things... we will skip most attributes
  switch( attr.name ) {
  case attATTACHMODIFYDATE:
    file.dt =  tnef_attr_parse_date( attr );
    break;

  case attATTACHMENT:
    var mapi_attrs = mapi_attr_read( attr.buf, attr.len );
    if( mapi_attrs ) {
      tnef_file_add_mapi_attrs( file, files, mapi_attrs );
      mapi_attr_free_list( mapi_attrs );
    }
    break;

  case attATTACHTITLE:
    file.name = tnef_file_munge_fname( attr.buf, files );
    break;

  case attATTACHDATA:
    file.len = attr.len;
    file.data = attr.buf;
    break;

  default:
    break;
  }
}

function tnef_file_clear( file ) {
  if( file.data )
    delete file.data;
}

// ---------- Tnef Date ----------

function tnef_date_parse( buf ) {
  var i = 0;
  var yr_num = GETINT16( buf, 0 );
  var mo_num = GETINT16( buf, 2 );
  var da_num = GETINT16( buf, 4 );
  var hr_num = GETINT16( buf, 6 );
  var mi_num = GETINT16( buf, 8 );
  var se_num = GETINT16( buf, 10 );
  return( new Date( yr_num, mo_num, da_num, hr_num, mi_num, se_num ) );
}


// ---------- Tnef Attr ----------

const MINIMUM_ATTR_LENGTH = 72;

// Object types
const LVL_MESSAGE    = 0x1;
const LVL_ATTACHMENT = 0x2;

function TRP() {
}
TRP.prototype = {
  id: 0,
  chbgtrp: 0,
  cch: 0,
  cb: 0
}

function TRIPLE() {
}
TRIPLE.prototype = {
  trp: TRP(),
  sender_display_name: null,
  sender_address: null
}


// Attr -- storing a structure, formated according to file specification
function TnefAttr() {
}
TnefAttr.prototype = {
  lvl_type: 0,
  type: 0,
  name: 0,
  len: 0,
  buf: null
}


// Copy the date data from the attribute into a struct date
function tnef_attr_parse_date( attr ) {
  assert( attr );
  assert( attr.type == szDATE );
  assert( attr.len >= 14 );

  return( tnef_date_parse( attr.buf ) );
}


function tnef_attr_parse_triple( attr, triple ) {
  assert( attr );
  assert( triple );
  assert( attr.type == szTRIPLES );

  triple.trp.id = GETINT16( attr.buf, 0 );
  triple.trp.chbgtrp = GETINT16( attr.buf, 2 );
  triple.trp.cch = GETINT16( attr.buf, 4 );
  triple.trp.cb = GETINT16( attr.buf, 6 );
  triple.sender_display_name = attr.buf[8];
  triple.sender_address = attr.buf[8] + triple.trp.cch;
}


// print attr to stderr.  Assumes that the Debug flag has been set and
// already checked
function tnef_attr_dbg_dump( attr ) {
  var name = get_tnef_name_str( attr.name );
  var type = get_tnef_type_str( attr.type );

  var msg = "(" + ((attr.lvl_type == LVL_MESSAGE) ? "MESS" : "ATTA") + ")" +
              name + "[type: " + type + "] [len: " + attr.len + "] =";

  switch( attr.type ) {
  case szBYTE:
    for( i = 0; i < attr.len; i++ )
      msg += " " + attr.buf.charCodeAt(i);
    break;

  case szSHORT:
    if( attr.len < 2 ) {
      msg += "Not enough data for szSHORT";
      dump( msg );
      return;
    }
    msg += " " + GETINT16( attr.buf );
    if( attr.len > 2 ) {
      msg += " [extra data:";
      for( i = 2; i < attr.len; i++ )
	msg += attr.buf.charCodeAt(i);
      msg += " ]";
    }
    break;

  case szLONG:
    if( attr.len < 4 ) {
      msg += "Not enough data for szLONG";
      dump( msg );
    }
    msg += " " + GETINT32( attr.buf );
    if( attr.len > 4 ) {
      msg += " [extra data:";
      for( var i = 4; i < attr.len; i++ )
	msg += " " + attr.buf.charCodeAt(i);
      msg += " ]";
    }
    break;

  case szWORD:
    for( var i = 0; i < attr.len; i += 2 )
      msg += " " + GETINT16( attr.buf, i );
    break;

  case szDWORD:
    for( var i = 0; i < attr.len; i += 4 )
      msg += " " + GETINT32( attr.buf, i );
    break;

  case szDATE:
    var dt = tnef_attr_copy_date_from_attr( attr );
    msg += dt.toString();
    dt = null;
    break;

  case szTEXT:
  case szSTRING:
    msg += attr.buf.substr( 0, attr.len - 1 );
    break;

  case szTRIPLES:
    var triple = tnef_attr_parse_triple( attr );
    msg += triple.toString();
    break;

  default:
    msg += "<unknown type>";
    break;
  }
  dump( msg + "\n" );
}

function tnef_attr_clear( attr ) {
  if( attr && attr.buf ) {
    delete attr.buf;
  }
}


// Validate the checksum against attr.  The checksum is the sum of all the
//   bytes in the attribute data modulo 65536
function tnef_attr_check_checksum( attr, checksum ) {
  var i, sum = 0;

  for( i = 0; i < attr.len; i++ )
    sum += attr.buf.charCodeAt(i);

  sum &= 0xFFFF;

  dump( "checksum = " + checksum + "  sum = " + sum );

  return( sum == checksum );
}


function tnef_attr_incomplete( attr ) {
  assert( attr );

  return( attr.len >= 0 && !(attr.buf && attr.buf.length == attr.len) );
}


function tnef_attr_read( instrm, prev_attr ) {
  var attr = null;

  if( prev_attr ) {
    attr = prev_attr;
  } else {
    attr = new TnefAttr();

    attr.lvl_type = instrm.read8( 1 );
    assert( attr.lvl_type == LVL_MESSAGE || attr.lvl_type == LVL_ATTACHMENT );

    //var type_and_name = GETINT32( instrm.readByteArray( 4 ) );
    attr.name = GETINT16( instrm.readByteArray( 2 ) ); //(((type_and_name << 16)&0xffffffff) >> 16)&0xffff;
    attr.type = GETINT16( instrm.readByteArray( 2 ) ); //(type_and_name >> 16)&0xffff;
    attr.len = GETINT32( instrm.readByteArray( 4 ) );
  }

  dump( "reading attr"
	+ "  lvl_type 0x" + to_hex( attr.lvl_type, 1 )
	+ "  name 0x" + to_hex( attr.name, 2 )
	+ "  type 0x" + to_hex( attr.type, 2 )
	+ "  length " + attr.len );

  var available = instrm.available();
  if( available > 0 ) {
    if( attr.buf ) {
      togo = attr.len - attr.buf.length;
    } else {
      togo = attr.len;
      attr.buf = "";
    }
    if( togo > available )
      togo = available;

    attr.buf += instrm.readBytes( togo );

    if( attr.buf.length == attr.len ) {
      checksum = GETINT16( instrm.readByteArray( 2 ) );
      if( !tnef_attr_check_checksum( attr, checksum ) ) {
	dump( "Invalid checksum, input file may be corrupted\n" );
	//exit( 1 );
      }
    } else {
      dump( "waiting for " + (attr.len - attr.buf.length) + " more bytes" );
    }
  }

  //if (DEBUG_ON) dbg_dump (attr);

  return( attr );
}


// ---------- MAPI -----------

const MULTI_VALUE_FLAG = 0x1000;
const GUID_EXISTS_FLAG = 0x8000;

function GUID() {
}
GUID.prototype = {
  data1: 0,
  data2: 0,
  data3: 0,
  data4: Array( 8 )
}

function MAPIValue() {
}
MAPIValue.prototype = {
  len: 0,
  data: null /* union {
    buf: Array(),
    bytes2: uint16,
    bytes4: uint32,
    bytes8: uint64,
    guid: GUID()
  } */
}

function MAPIAttr() {
}
MAPIAttr.prototype = {
  type: 0,
  name: 0,
  num_values: 0,
  values: null,
  guid: null,
  num_names: 0,
  names: null
}

function pad_to_4byte( x ) {
  return( Math.ceil( x/4.0 )*4 );
}

function guid_parse_buf( buf, idx ) {
  var guid = new GUID();

  assert( buf );

  guid.data1 = GETINT32( buf, idx );
  guid.data1 = GETINT16( buf, idx + 4 );
  guid.data1 = GETINT16( buf, idx + 6 );
  for( var i = 0; i < 8; i++ )
    guid.data4[i] = buf[idx + 8 + i];

  return( guid );
}

// parses out the MAPI attibutes hidden in the character buffer
function mapi_attr_read( buf, len ) {
  var idx = 0;
  var i = 0;
  var num_properties = GETINT32( buf, idx );
  var mattrs = Array( num_properties + 1 );

  idx += 4;

  for( i = 0; i < num_properties; i++ ) {
    mattrs[i] = new MAPIAttr();

    mattrs[i].type = GETINT16( buf, idx );
    idx += 2;
    mattrs[i].name = GETINT16( buf, idx );
    idx += 2;

    // Multi-valued attributes have their type modified by the MULTI_VALUE_FLAG value
    if( mattrs[i].type & MULTI_VALUE_FLAG )
      mattrs[i].type -= MULTI_VALUE_FLAG;

    // handle special case of GUID prefixed properties
    if( mattrs[i].name >= GUID_EXISTS_FLAG ) {
      // copy GUID
      mattrs[i].guid = guid_parse_buf( buf, idx );
      idx += 16;

      mattrs[i].num_names = GETINT32( buf, idx );
      idx += 4;
      if( mattrs[i].num_names > 0 ) {
	// FIXME: do something useful here!
	var i2 = 0;

	mattrs[i].names = Array( mattrs[i].num_names );

	for( i2 = 0; i2 < attrs[i].num_names; i2++ ) {
	  var name_len = GETINT32( buf, idx );
	  idx += 4;

	  // read the data into a buffer
	  mattrs[i].names[i2] = buf.substr( idx, name_len );

	  // But what are we going to do with it?

	  idx += pad_to_4byte( name_len );
	}
      } else {
	// get the 'real' name
	mattrs[i].name = GETINT32( buf, idx );
	idx += 4;
      }
    }

    switch( mattrs[i].type ) {
    case szMAPI_NULL:
      mattrs[i].num_values = 0;
      mattrs[i].values = null;
      break;

    case szMAPI_SHORT:        // 2 bytes
      mattrs[i].num_values = 1;
      mattrs[i].values = [ GETINT16( buf, idx ) ];
      idx += 4; // advance by 4!
      break;

    case szMAPI_INT:
    case szMAPI_FLOAT:      // 4 bytes
    case szMAPI_ERROR:
    case szMAPI_BOOLEAN:	// spec says 2 bytes but reality is 4!
      mattrs[i].num_values = 1;
      mattrs[i].values = [ GETINT32( buf, idx ) ];
      idx += 4;
      break;

    case szMAPI_DOUBLE:
    case szMAPI_APPTIME:
    case szMAPI_CURRENCY:
    case szMAPI_INT8BYTE:
    case szMAPI_SYSTIME:         // 8 bytes
      mattrs[i].num_values = 1;
      mattrs[i].values = [ GETINT64( buf, idx ) ];
      idx += 8;
      break;

    case szMAPI_CLSID:
      mattrs[i].num_values = 1;
      mattrs[i].values = [ guid_parse_buf( buf, idx ) ];
      idx += 16; //sizeof (GUID);
      break;

    case szMAPI_STRING:
    case szMAPI_UNICODE_STRING:
    case szMAPI_OBJECT:
    case szMAPI_BINARY:       // variable length
    case szMAPI_UNSPECIFIED:
      mattrs[i].num_values = GETINT32( buf, idx );
      idx += 4;
      mattrs[i].values = new Array( mattrs[i].num_values );
      for( var val_idx = 0; val_idx < mattrs[i].num_values; val_idx++ ) {
	var val_len = GETINT32( buf, idx );
	idx += 4;

	if( mattrs[i].name == szMAPI_STRING )
	  val_len--; // kill c staring terminator

	if( mattrs[i].type == szMAPI_UNICODE_STRING ) {
	  mattrs[i].values[val_idx] = unicode_to_utf8( buf.substr( idx, val_len ) );
	} else {
	  mattrs[i].values[val_idx] = buf.substr( idx, val_len - 1 );
	}
	idx += pad_to_4byte( val_len );
      }
      break;
    }
    //if (DEBUG_ON) mapi_attr_dump (a);
  }
  mattrs[i] = null;

  return( mattrs );
}

function mapi_attr_clear( mattr ) {
  if( mattr ) {
    var i = 0;
    for( i = 0; i < mattr.num_values; i++ ) {
      if( mattr.type == szMAPI_STRING ||
	  mattr.type == szMAPI_UNICODE_STRING ||
	  mattr.type == szMAPI_BINARY ) {
	mattr.values[i] = null;
      }
    }
    mattr.values = null;
    mattr.guid = null;
  }
}

function mapi_attr_free_list( mattrs ) {
  var i;

  for( i = 0; mattrs && mattrs[i]; i++ ) {
    mapi_attr_clear( mattrs[i] );
    mattrs[i] = null;
  }
}


// ---------- RTF -----------

const rtf_uncompressed_magic = 0x414c454d;
const rtf_compressed_magic = 0x75465a4c;
const rtf_prebuf = "{\\rtf1\\ansi\\mac\\deff0\\deftab720{\\fonttbl;}{\\f0\\fnil \\froman \\fswiss \\fmodern \\fscript \\fdecor MS Sans SerifSymbolArialTimes New RomanCourier{\\colortbl\\red0\\green0\\blue0\n\r\\par \\pard\\plain\\f0\\fs20\\b\\i\\u\\tab\\tx";


function is_rtf_data( data ) {
  var compr_size = 0;
  var uncompr_size = 0;
  var magic;
  var idx = 0;

  compr_size = GETINT32( data, idx );
  idx += 4;
  uncompr_size = GETINT32( data, idx );
  idx += 4;
  magic = GETINT32( data, idx );
  idx += 4;

  if( magic == rtf_uncompressed_magic || magic == rtf_compressed_magic )
    return 1;
  return 0;
}


function decompress_rtf_data( src, len ) {
  const rtf_prebuf_len = rtf_prebuf.length;
  var numin = 0;
  var out = 0;
  var flag_count = 0;
  var flags = 0;
  var dest = new String( "" ); // size is rtf_prebuf_len + len;
  var i;

  for( i = 0; i < rtf_prebuf.length; i++ )
    dest[i] = rtf_prebuf.charCodeAt(i);

  out = rtf_prebuf_len;

  while( out < len + rtf_prebuf_len ) {
    // each flag byte flags 8 literals/references, 1 per bit
    flags = (((flag_count++ % 8) == 0) ? src.charCodeAt(numin++) : flags >> 1);
    if( (flags & 1) == 1 ) {	// 1 == reference
      var offset = src.charCodeAt( numin );
      var length = src.charCodeAt( numin + 1 );
      numin += 2;

      // offset relative to block start
      offset = (offset << 4) | (length >> 4);

      // number of bytes to copy
      length = (length & 0xF) + 2;

      /* decompression buffer is supposed to wrap around back to
       * the beginning when the end is reached.  we save the
       * need for this by pointing straight into the data
       * buffer, and simulating this behaviour by modifying the
       * pointers appropriately */ 
      offset = (out / 4096) * 4096 + offset;
      if( offset >= out )
	offset -= 4096; // from previous block

      var end = offset + length;
      while( offset < end ) {
	dest += dest[offset];
	offset++;
      }
    } else {		// 0 == literal
      dest += src[numin++];
    }
  }
    
  return( dest.slice( rtf_prebuf_len, rtf_prebuf_len + len - 1 ) );
}

function rtf_data_parse( data, len ) {
  var out_data = null;
  var out_len = 0;
  var compr_size = 0;
  var uncompr_size = 0;
  var magic;
  var checksum;
  var idx = 0;

  compr_size = GETINT32( data, idx );
  idx += 4;
  uncompr_size = GETINT32( data, idx );
  idx += 4;
  magic = GETINT32( data, idx );
  idx += 4;
  checksum = GETINT32( data, idx );
  idx += 4;

  return( null );
  // sanity check
  assert( compr_size + 4 == len );
  if( compr_size + 4 != len ) {
    dump( "size does not match: got " + (compr_size + 4) + ", expected: " + len );
    return null;
  }
  out_len = uncompr_size;

  if( magic == rtf_uncompressed_magic ) // uncompressed rtf stream
    out_data = data.slice( 4, 4 + uncompr_size - 1 );
  else if( magic == rtf_compressed_magic ) // compressed rtf stream
    out_data = decompress_rtf_data( data.slice( idx, idx + compr_size - 1 ), uncompr_size );

  return( out_data );
}



// ---------- Tnef -----------

//static size_t filesize;

/* TNEF signature.  Equivalent to the magic cookie for a TNEF file. */
const TNEF_SIGNATURE = 0x223e9f78;


function TnefListenerInterface() {
}
TnefListenerInterface.prototype = {  
  onTnefStart: function ( filename, content_type, length, date ) { },
  onTnefData: function ( position, data ) { },
  onTnefEnd: function ( ) { }
}

function TnefPackage() {
}
TnefPackage.prototype = {
  files: Array(),

  cur_file: null,
  cur_attr: null,
}

//MessageBodyTypes {
const MSG_TEXT = 't';
const MSG_HTML = 'h';
const MSG_RTF  = 'r';
//}

// Reads and decodes a object from the stream
function tnef_pack_read_object( pkg, instrm ) {
  var attr = null;

  // peek to see if there is more to read from this stream
  if( instrm.available() < MINIMUM_ATTR_LENGTH )
    return( null );

  attr = tnef_attr_read( instrm, pkg.cur_attr );

  return( attr );
}


function tnef_pack_handle_body_part( pkg, data, len, content_type, listener ) {
  if( !data )
    return;

  var file = new TnefFile( pkg );

  if( typeof( data ) == "Array" )
    file.data = String.fromCharCode.apply( null, data );
  else
    file.data = data;

  file.name = "";
  file.len = len;
  file.mime_type = content_type;
  tnef_file_notify( file, listener, true );
}

function tnef_pack_handle_text_attr( pkg, attr, listener ) {
  tnef_pack_handle_body_part( pkg, attr.buf, attr.len, "text/plain", listener );
}

function tnef_pack_handle_html_data( pkg, mattr, listener ) {
  for( var j = 0; j < mattr.num_values; j++ )
    tnef_pack_handle_body_part( pkg, mattr.values[j], mattr.values[j].length, "text/html", listener );
}


function tnef_pack_handle_rtf_data( pkg, mattr, listener ) {
  for( var j = 0; j < mattr.num_values; j++ ) {
    if( is_rtf_data( mattr.values[j] ) ) {
      var rtf_data = rtf_data_parse( mattr.values[j], mattr.values[j].length );
      if( rtf_data )
	tnef_pack_handle_body_part( pkg, rtf_data, rtf_data.length, "application/rtf", listener );
    }
  }
}

function tnef_pack_data_left( instrm ) {
  var dl = instrm.available();

  if( dl > 0 && dl < MINIMUM_ATTR_LENGTH ) {
    dump( "ERROR: garbage at end of file.\n" );
    return( false );
  }

  return( true );
}


// The entry point into this module.  This parses an entire TNEF file.
function tnef_pack_parse_stream( instrm, listener, prev_pack ) {
  var sig = 0;
  var key = 0;
  var pkg = null;

  if( prev_pack ) {
    pkg = prev_pack;
  } else {
    pkg = new TnefPackage();

    // check that this is in fact a TNEF file
    sig = GETINT32( instrm.readByteArray( 4 ) );
    if( sig != TNEF_SIGNATURE ) {
      dump( "sig = " + sig );
      alert( "Seems not to be a TNEF file\n" );
      return( null );
    }

    // Get the key
    key = GETINT16( instrm.readByteArray( 2 ) );
    //debug_print( "TNEF Key: %hx\n", key );
  }

  // The rest of the file is a series of 'messages' and 'attachments'
  pkg.cur_attr = tnef_pack_read_object( pkg, instrm );
  if( pkg.cur_attr && tnef_attr_incomplete( pkg.cur_attr ) )
    return( pkg );
  while( pkg.cur_attr && tnef_pack_data_left( instrm ) ) {
    // This signals the beginning of a file
    if( pkg.cur_attr.name == attATTACHRENDDATA ) {
      if( pkg.cur_file ) {
	tnef_file_notify( pkg.cur_file, listener, true );
	tnef_file_clear( pkg.cur_file );
      } else {
	pkg.cur_file = new TnefFile( pkg );
	dump( "starting file sub-attachment" );
      }
    }
    // Add the data to our lists.
    switch( pkg.cur_attr.lvl_type ) {
    case LVL_MESSAGE:
      if( pkg.cur_attr.name == attBODY ) {
	tnef_pack_handle_text_attr( pkg, pkg.cur_attr, listener );
      } else if( pkg.cur_attr.name == attMAPIPROPS ) { 
	var mapi_attrs = mapi_attr_read( pkg.cur_attr.buf, pkg.cur_attr.len ); 

	if( mapi_attrs ) {
	  for( i = 0; mapi_attrs[i]; i++ ) {
	    switch( mapi_attrs[i].name ) {
	    case MAPI_BODY_HTML:
	      tnef_pack_handle_html_data( pkg, mapi_attrs[i], listener );
	      break;
	    case MAPI_RTF_COMPRESSED:
	      tnef_pack_handle_rtf_data( pkg, mapi_attrs[i], listener );
	      break;
	    default:
	      // cannot save attributes to file yet, since they are not attachment attributes
	      // tnef_file_add_mapi_attrs( pkg.cur_file, mapi_attrs );
	      break;
	    }
	  }
	  mapi_attrs = null;
	}
      }
      break;
    case LVL_ATTACHMENT:
      if( !pkg.cur_file )
	pkg.cur_file = new TnefFile( pkg );
      tnef_file_add_attr( pkg.cur_file, pkg.files, pkg.cur_attr );

      // remember attachment file name to avoid collisions
      if( pkg.cur_file.name && pkg.cur_file.len > 0 )
	pkg.files.push( pkg.cur_file.name );
      break;
    default:
      dump( "Invalid lvl type on attribute: " + pkg.cur_attr.lvl_type );
      return( 1 );
      break;
    }

    pkg.cur_attr = null;
    pkg.cur_attr = tnef_pack_read_object( pkg, instrm );
    if( pkg.cur_attr && tnef_attr_incomplete( pkg.cur_attr ) )
      return( pkg );
  }
  if( pkg.cur_file ) {
    tnef_file_notify( pkg.cur_file, listener, true );
    tnef_file_clear( pkg.cur_file );
    pkg.cur_file = null;
  }
  if( pkg.cur_attr ) {
    pkg.cur_attr = null;
  }

  return( null );
}

